#
# This file is part of pySMT.
#
#   Copyright 2014 Andrea Micheli and Marco Gario
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import atexit
from warnings import warn

from pysmt.exceptions import SolverAPINotFound

try:
    import yicespy
except ImportError:
    raise SolverAPINotFound


from pysmt.solvers.eager import EagerModel
from pysmt.solvers.solver import Solver, Converter, SolverOptions
from pysmt.solvers.smtlib import SmtLibBasicSolver, SmtLibIgnoreMixin

from pysmt.walkers import DagWalker
from pysmt.exceptions import SolverReturnedUnknownResultError
from pysmt.exceptions import (
    InternalSolverError,
    NonLinearError,
    PysmtValueError,
    PysmtTypeError,
)
from pysmt.decorators import clear_pending_pop, catch_conversion_error
from pysmt.constants import Fraction, is_pysmt_integer

import pysmt.logics


# Initialization
def init():
    if not getattr(init, "initialized", False):
        yicespy.yices_init()
    init.initialized = True


def reset_yices():
    yicespy.yices_reset()


init()


@atexit.register
def cleanup():
    yicespy.yices_exit()


# Yices constants
STATUS_UNKNOWN = 2
STATUS_SAT = 3
STATUS_UNSAT = 4


def yices_logic(pysmt_logic):
    """Return a Yices String representing the given pySMT logic."""
    ylogic = str(pysmt_logic)
    if ylogic == "QF_BOOL":
        ylogic = "NONE"
    return ylogic


class YicesOptions(SolverOptions):
    def __init__(self, **base_options):
        SolverOptions.__init__(self, **base_options)
        # TODO: Yices Supports UnsatCore extraction
        # but we did not wrapped it yet.
        if self.unsat_cores_mode is not None:
            raise PysmtValueError("'unsat_cores_mode' option not supported.")

    @staticmethod
    def _set_option(cfg, name, value):
        rv = yicespy.yices_set_config(cfg, name, value)
        if rv != 0:
            # This might be a parameter to be set later (see set_params)
            # We raise the exception only if the parameter exists but the value
            # provided to the parameter is invalid.
            err = yicespy.yices_error_code()
            if err == yicespy.CTX_INVALID_PARAMETER_VALUE:
                raise PysmtValueError(
                    "Error setting the option " "'%s=%s'" % (name, value)
                )

    def __call__(self, solver):
        if self.generate_models:
            # Yices always generates models
            pass
        if self.incremental:
            self._set_option(solver.yices_config, "mode", "push-pop")
        else:
            self._set_option(solver.yices_config, "mode", "one-shot")

        if self.random_seed is not None:
            self._set_option(solver.yices_config, "random-seed", str(self.random_seed))

        for k, v in self.solver_options.items():
            self._set_option(solver.yices_config, str(k), str(v))

    def set_params(self, solver):
        """Set Search Parameters.

        Yices makes a distinction between configuratin and search
        parameters.  The first are fixed for the lifetime of a
        context, while the latter can be different for every call to
        check_context.

        A list of available parameters is available at:
        http://yices.csl.sri.com/doc/parameters.html
        """
        params = yicespy.yices_new_param_record()
        yicespy.yices_default_params_for_context(solver.yices, params)
        for k, v in self.solver_options.items():
            rv = yicespy.yices_set_param(params, k, v)
            if rv != 0:
                raise PysmtValueError("Error setting the option '%s=%s'" % (k, v))
        solver.yices_params = params


# EOC YicesOptions


class YicesSolver(Solver, SmtLibBasicSolver, SmtLibIgnoreMixin):

    LOGICS = (
        pysmt.logics.PYSMT_QF_LOGICS
        - pysmt.logics.ARRAYS_LOGICS
        - set(
            l
            for l in pysmt.logics.PYSMT_QF_LOGICS
            if not l.theory.linear or l.theory.strings
        )
    )
    OptionsClass = YicesOptions

    def __init__(self, environment, logic, **options):
        Solver.__init__(self, environment=environment, logic=logic, **options)

        self.declarations = set()
        self.yices_config = yicespy.yices_new_config()
        if (
            yicespy.yices_default_config_for_logic(
                self.yices_config, yices_logic(logic)
            )
            != 0
        ):
            warn("Error setting config for logic %s" % logic)
        self.options(self)
        self.yices = yicespy.yices_new_context(self.yices_config)
        self.options.set_params(self)
        yicespy.yices_free_config(self.yices_config)
        self.converter = YicesConverter(environment)
        self.mgr = environment.formula_manager
        self.model = None
        self.failed_pushes = 0
        return

    @clear_pending_pop
    def reset_assertions(self):
        yicespy.yices_reset_context(self.yices)

    @clear_pending_pop
    def declare_variable(self, var):
        raise NotImplementedError

    @clear_pending_pop
    def add_assertion(self, formula, named=None):
        self._assert_is_boolean(formula)
        term = self.converter.convert(formula)
        code = yicespy.yices_assert_formula(self.yices, term)
        if code != 0:
            msg = yicespy.yices_error_string()
            if code == -1 and "non-linear arithmetic" in msg:
                raise NonLinearError(formula)
            raise InternalSolverError(
                "Yices returned non-zero code upon assert"
                ": %s (code: %s)" % (msg, code)
            )

    def get_model(self):
        assignment = {}
        # MG: This iteration is problematic, since it assumes that all
        # defined symbols have a type that is compatible with this
        # solver.  In this case, the problem occurs with Arrays and
        # Strings that are not supported.
        for s in self.environment.formula_manager.get_all_symbols():
            if s.is_symbol() and s.symbol_type().is_string_type():
                continue
            if s.is_term():
                if s.symbol_type().is_array_type():
                    continue
                if s.symbol_type().is_custom_type():
                    continue
                v = self.get_value(s)
                assignment[s] = v
        return EagerModel(assignment=assignment, environment=self.environment)

    @clear_pending_pop
    def solve(self, assumptions=None):
        if assumptions is not None:
            self.push()
            self.add_assertion(self.mgr.And(assumptions))
            self.pending_pop = True

        out = yicespy.yices_check_context(self.yices, self.yices_params)

        if self.model is not None:
            yicespy.yices_free_model(self.model)
            self.model = None

        assert out in [STATUS_SAT, STATUS_UNSAT, STATUS_UNKNOWN]
        if out == STATUS_UNKNOWN:
            raise SolverReturnedUnknownResultError()
        elif out == STATUS_SAT:
            self.model = yicespy.yices_get_model(self.yices, 1)
            return True
        else:
            return False

    @clear_pending_pop
    def all_sat(self, important, callback):
        raise NotImplementedError

    @clear_pending_pop
    def push(self, levels=1):
        for _ in range(levels):
            c = yicespy.yices_push(self.yices)
            if c != 0:
                # 4 is STATUS_UNSAT
                if yicespy.yices_context_status(self.yices) == 4:
                    # Yices fails to push if the context is in UNSAT state
                    # (It makes no sense to conjoin formulae to False)
                    # PySMT allows this and we support it by counting the
                    # spurious push calls
                    self.failed_pushes += 1
                else:
                    raise InternalSolverError(
                        "Error in push: %s" % yicespy.yices_error_string()
                    )

    @clear_pending_pop
    def pop(self, levels=1):
        for _ in range(levels):
            if self.failed_pushes > 0:
                self.failed_pushes -= 1
            else:
                c = yicespy.yices_pop(self.yices)
                if c != 0:
                    raise InternalSolverError(
                        "Error in pop: %s" % yicespy.yices_error_string()
                    )

    def print_model(self, name_filter=None):
        for var in self.declarations:
            if name_filter is None or not var.symbol_name().startswith(name_filter):
                print("%s = %s", (var.symbol_name(), self.get_value(var)))

    def _check_error(self, res):
        if res != 0:
            err = yicespy.yices_error_string()
            raise InternalSolverError("Yices returned an error: " + err)

    def get_value(self, item):
        self._assert_no_function_type(item)

        titem = self.converter.convert(item)
        ty = self.environment.stc.get_type(item)
        if ty.is_bool_type():
            status, res = yicespy.yices_get_bool_value(self.model, titem)
            self._check_error(status)
            return self.mgr.Bool(bool(res))
        elif ty.is_int_type():
            res = yicespy.yices_get_integer_value(self.model, titem)
            return self.mgr.Int(res)
        elif ty.is_real_type():
            status, val = yicespy.yices_get_rational_value(self.model, titem)
            self._check_error(status)
            return self.mgr.Real(Fraction(val))
        elif ty.is_bv_type():
            status, res = yicespy.yices_get_bv_value(self.model, titem, ty.width)
            self._check_error(status)
            str_val = "".join(str(x) for x in reversed(res))
            return self.mgr.BV("#b" + str_val)
        else:
            raise NotImplementedError()

    def _exit(self):
        yicespy.yices_free_context(self.yices)
        yicespy.yices_free_param_record(self.yices_params)


# EOC YicesSolver


class YicesConverter(Converter, DagWalker):
    def __init__(self, environment):
        DagWalker.__init__(self, environment)
        self.backconversion = {}
        self.mgr = environment.formula_manager
        self._get_type = environment.stc.get_type

        # Maps a Symbol into the corresponding internal yices instance
        self.symbol_to_decl = {}
        # Maps an internal yices instance into the corresponding symbol
        self.decl_to_symbol = {}
        self._yicesSort = {}

    @catch_conversion_error
    def convert(self, formula):
        return self.walk(formula)

    def _check_term_result(self, res):
        if res == -1:
            err = yicespy.yices_error_string()
            raise InternalSolverError("Yices returned an error: " + err)

    def walk_and(self, formula, args, **kwargs):
        res = yicespy.yices_and(len(args), args)
        self._check_term_result(res)
        return res

    def walk_or(self, formula, args, **kwargs):
        res = yicespy.yices_or(len(args), args)
        self._check_term_result(res)
        return res

    def walk_not(self, formula, args, **kwargs):
        res = yicespy.yices_not(args[0])
        self._check_term_result(res)
        return res

    def walk_symbol(self, formula, **kwargs):
        symbol_type = formula.symbol_type()
        var_type = self._type_to_yices(symbol_type)
        term = yicespy.yices_new_uninterpreted_term(var_type)
        yicespy.yices_set_term_name(term, formula.symbol_name())
        self._check_term_result(term)
        return term

    def _bound_symbol(self, var):
        symbol_type = var.symbol_type()
        var_type = self._type_to_yices(symbol_type)
        term = yicespy.yices_new_variable(var_type)
        yicespy.yices_set_term_name(term, var.symbol_name())
        return term

    def walk_iff(self, formula, args, **kwargs):
        res = yicespy.yices_iff(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_implies(self, formula, args, **kwargs):
        res = yicespy.yices_implies(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_le(self, formula, args, **kwargs):
        res = yicespy.yices_arith_leq_atom(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_lt(self, formula, args, **kwargs):
        res = yicespy.yices_arith_lt_atom(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_ite(self, formula, args, **kwargs):
        i, t, e = args
        res = yicespy.yices_ite(i, t, e)
        self._check_term_result(res)
        return res

    def walk_real_constant(self, formula, **kwargs):
        frac = formula.constant_value()
        n, d = frac.numerator, frac.denominator
        rep = str(n) + "/" + str(d)
        res = yicespy.yices_parse_rational(rep)
        self._check_term_result(res)
        return res

    def walk_int_constant(self, formula, **kwargs):
        assert is_pysmt_integer(formula.constant_value())
        rep = str(formula.constant_value())
        res = yicespy.yices_parse_rational(rep)
        self._check_term_result(res)
        return res

    def walk_bool_constant(self, formula, **kwargs):
        if formula.constant_value():
            return yicespy.yices_true()
        else:
            return yicespy.yices_false()

    def walk_exists(self, formula, args, **kwargs):
        (bound_formula, var_list) = self._rename_bound_variables(
            args[0], formula.quantifier_vars()
        )
        res = yicespy.yices_exists(len(var_list), var_list, bound_formula)
        self._check_term_result(res)
        return res

    def walk_forall(self, formula, args, **kwargs):
        (bound_formula, var_list) = self._rename_bound_variables(
            args[0], formula.quantifier_vars()
        )
        res = yicespy.yices_forall(len(var_list), var_list, bound_formula)
        self._check_term_result(res)
        return res

    def _rename_bound_variables(self, formula, variables):
        """Bounds the variables in formula.

        Returns a tuple (new_formula, new_var_list) in which the old
        variables have been replaced by the new variables in the list.
        """
        new_vars = [self._bound_symbol(x) for x in variables]
        old_vars = [self.walk_symbol(x) for x in variables]
        new_formula = yicespy.yices_subst_term(
            len(variables), new_vars, old_vars, formula
        )
        return (new_formula, new_vars)

    def walk_plus(self, formula, args, **kwargs):
        res = yicespy.yices_sum(len(args), args)
        self._check_term_result(res)
        return res

    def walk_minus(self, formula, args, **kwargs):
        res = yicespy.yices_sub(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_equals(self, formula, args, **kwargs):
        tp = self._get_type(formula.arg(0))
        res = None
        if tp.is_bv_type():
            res = yicespy.yices_bveq_atom(args[0], args[1])
        elif tp.is_int_type() or tp.is_real_type():
            res = yicespy.yices_arith_eq_atom(args[0], args[1])
        else:
            assert tp.is_custom_type()
            res = yicespy.yices_eq(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_times(self, formula, args, **kwargs):
        res = args[0]
        for x in args[1:]:
            res = yicespy.yices_mul(res, x)
            self._check_term_result(res)
        return res

    def walk_toreal(self, formula, args, **kwargs):
        return args[0]

    def walk_function(self, formula, args, **kwargs):
        name = formula.function_name()
        if name not in self.symbol_to_decl:
            self.declare_variable(name)
        decl = self.symbol_to_decl[name]
        res = yicespy.yices_application(decl, len(args), args)
        self._check_term_result(res)
        return res

    def walk_bv_constant(self, formula, **kwargs):
        width = formula.bv_width()
        res = None
        value = formula.constant_value()
        if value <= ((2**63) - 1):
            # we can use the numerical representation
            # Note: yicespy uses *signed* longs in the API, so the maximal
            # representable number is 2^63 - 1
            res = yicespy.yices_bvconst_uint64(width, value)
        else:
            # we must resort to strings to communicate the result to yices
            res = yicespy.yices_parse_bvbin(formula.bv_bin_str())
        self._check_term_result(res)
        return res

    def walk_bv_ult(self, formula, args, **kwargs):
        res = yicespy.yices_bvlt_atom(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_bv_ule(self, formula, args, **kwargs):
        res = yicespy.yices_bvle_atom(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_bv_concat(self, formula, args, **kwargs):
        res = yicespy.yices_bvconcat2(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_bv_extract(self, formula, args, **kwargs):
        res = yicespy.yices_bvextract(
            args[0], formula.bv_extract_start(), formula.bv_extract_end()
        )
        self._check_term_result(res)
        return res

    def walk_bv_or(self, formula, args, **kwargs):
        res = yicespy.yices_bvor2(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_bv_not(self, formula, args, **kwargs):
        res = yicespy.yices_bvnot(args[0])
        self._check_term_result(res)
        return res

    def walk_bv_and(self, formula, args, **kwargs):
        res = yicespy.yices_bvand2(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_bv_xor(self, formula, args, **kwargs):
        res = yicespy.yices_bvxor2(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_bv_add(self, formula, args, **kwargs):
        res = yicespy.yices_bvadd(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_bv_sub(self, formula, args, **kwargs):
        res = yicespy.yices_bvsub(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_bv_neg(self, formula, args, **kwargs):
        res = yicespy.yices_bvneg(args[0])
        self._check_term_result(res)
        return res

    def walk_bv_mul(self, formula, args, **kwargs):
        res = yicespy.yices_bvmul(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_bv_udiv(self, formula, args, **kwargs):
        res = yicespy.yices_bvdiv(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_bv_urem(self, formula, args, **kwargs):
        res = yicespy.yices_bvrem(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_bv_lshl(self, formula, args, **kwargs):
        res = yicespy.yices_bvshl(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_bv_lshr(self, formula, args, **kwargs):
        res = yicespy.yices_bvlshr(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_bv_rol(self, formula, args, **kwargs):
        res = yicespy.yices_rotate_left(args[0], formula.bv_rotation_step())
        self._check_term_result(res)
        return res

    def walk_bv_ror(self, formula, args, **kwargs):
        res = yicespy.yices_rotate_right(args[0], formula.bv_rotation_step())
        self._check_term_result(res)
        return res

    def walk_bv_zext(self, formula, args, **kwargs):
        res = yicespy.yices_zero_extend(args[0], formula.bv_extend_step())
        self._check_term_result(res)
        return res

    def walk_bv_sext(self, formula, args, **kwargs):
        res = yicespy.yices_sign_extend(args[0], formula.bv_extend_step())
        self._check_term_result(res)
        return res

    def walk_bv_slt(self, formula, args, **kwargs):
        res = yicespy.yices_bvslt_atom(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_bv_sle(self, formula, args, **kwargs):
        res = yicespy.yices_bvsle_atom(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_bv_comp(self, formula, args, **kwargs):
        a, b = args
        eq = yicespy.yices_bveq_atom(a, b)
        self._check_term_result(eq)
        one = yicespy.yices_bvconst_int32(1, 1)
        zero = yicespy.yices_bvconst_int32(1, 0)
        res = yicespy.yices_ite(eq, one, zero)
        self._check_term_result(res)
        return res

    def walk_bv_sdiv(self, formula, args, **kwargs):
        res = yicespy.yices_bvsdiv(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_bv_srem(self, formula, args, **kwargs):
        res = yicespy.yices_bvsrem(args[0], args[1])
        self._check_term_result(res)
        return res

    def walk_bv_ashr(self, formula, args, **kwargs):
        res = yicespy.yices_bvashr(args[0], args[1])
        self._check_term_result(res)
        return res

    def yicesSort(self, name):
        """Return the yices Sort for the given name."""
        name = str(name)
        try:
            return self._yicesSort[name]
        except KeyError:
            sort = yicespy.yices_new_uninterpreted_type()
            self._yicesSort[name] = sort
        return sort

    def _type_to_yices(self, tp):
        if tp.is_bool_type():
            return yicespy.yices_bool_type()
        elif tp.is_real_type():
            return yicespy.yices_real_type()
        elif tp.is_int_type():
            return yicespy.yices_int_type()
        elif tp.is_function_type():
            stps = [self._type_to_yices(x) for x in tp.param_types]
            rtp = self._type_to_yices(tp.return_type)
            # arr = (yicespy.type_t * len(stps))(*stps)
            return yicespy.yices_function_type(len(stps), stps, rtp)
        elif tp.is_bv_type():
            return yicespy.yices_bv_type(tp.width)
        elif tp.is_custom_type():
            return self.yicesSort(str(tp))
        else:
            raise NotImplementedError(tp)

    def declare_variable(self, var):
        if not var.is_symbol():
            raise PysmtTypeError(
                "Trying to declare as a variable something "
                "that is not a symbol: %s" % var
            )
        if var.symbol_name() not in self.symbol_to_decl:
            tp = self._type_to_yices(var.symbol_type())
            decl = yicespy.yices_new_uninterpreted_term(tp)
            yicespy.yices_set_term_name(decl, var.symbol_name())
            self.symbol_to_decl[var] = decl
            self.decl_to_symbol[decl] = var



#
# This file is part of pySMT.
#
#   Copyright 2014 Andrea Micheli and Marco Gario
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import pytest

from pysmt.test import TestCase, skipIfSolverNotAvailable, main
from pysmt.test.examples import get_example_formulae
from pysmt.environment import get_env
from pysmt.shortcuts import (
    Array,
    Store,
    Int,
    Iff,
    Symbol,
    Plus,
    Equals,
    And,
    Real,
    Times,
    Not,
    FALSE,
    Or,
    TRUE,
)
from pysmt.typing import INT, REAL
from pysmt.simplifier import BddSimplifier
from pysmt.logics import QF_BOOL
from pysmt.exceptions import ConvertExpressionError, NoSolverAvailableError


class TestSimplify(TestCase):
    @pytest.mark.slow
    @skipIfSolverNotAvailable("z3")
    @skipIfSolverNotAvailable("cvc4")
    def test_simplify_qf(self):
        simp = get_env().simplifier
        for (f, _, _, logic) in get_example_formulae():
            if logic.is_quantified():
                continue
            sname = "z3" if not logic.theory.strings else "cvc4"
            simp.validate_simplifications = sname
            sf = f.simplify()
            simp.validate_simplifications = None
            self.assertValid(
                Iff(f, sf),
                solver_name=sname,
                msg="Simplification did not provide equivalent "
                + "result:\n f= %s\n sf = %s" % (f, sf),
            )

    @pytest.mark.slow
    @skipIfSolverNotAvailable("z3")
    def test_simplify_q(self):
        simp = get_env().simplifier
        for (f, _, _, logic) in get_example_formulae():
            if logic.quantifier_free:
                continue
            simp.validate_simplifications = "z3"
            sf = f.simplify()
            simp.validate_simplifications = None
            self.assertValid(
                Iff(f, sf),
                solver_name="z3",
                msg="Simplification did not provide equivalent "
                + "result:\n f= %s\n sf = %s" % (f, sf),
            )

    def test_array_value(self):
        a1 = Array(INT, Int(0), {Int(12): Int(10)})
        a2 = Store(Array(INT, Int(0)), Int(12), Int(10))
        self.assertEqual(a1, a2.simplify())

    @skipIfSolverNotAvailable("bdd")
    def test_bdd_simplify(self):
        s = BddSimplifier()
        for (f, _, _, logic) in get_example_formulae():
            if logic == QF_BOOL:
                fprime = s.simplify(f)
                self.assertValid(Iff(fprime, f))

        s = BddSimplifier()
        try:
            s.validate_simplifications = True
        except ValueError:
            self.assertTrue(len(self.env.factory.all_solvers()) == 1)
        for (f, _, _, logic) in get_example_formulae():
            if logic == QF_BOOL:
                fprime = s.simplify(f)
                self.assertValid(Iff(fprime, f))

    @skipIfSolverNotAvailable("bdd")
    @skipIfSolverNotAvailable("z3")
    def test_bdd_simplify_bool_abs(self):
        s = BddSimplifier()
        for (f, _, _, logic) in get_example_formulae():
            if not logic.theory.linear:
                continue
            if logic != QF_BOOL:
                with self.assertRaises(ConvertExpressionError):
                    s.simplify(f)

        s = BddSimplifier(bool_abstraction=True)
        for (f, _, _, logic) in get_example_formulae():
            if logic.quantifier_free:
                fprime = s.simplify(f)
                try:
                    self.assertValid(Iff(fprime, f))
                except NoSolverAvailableError:
                    pass

        s = BddSimplifier(bool_abstraction=True)
        f = And(Equals(Plus(Int(5), Int(1)), Int(6)), Symbol("x"))
        fs = s.simplify(f)
        self.assertEqual(fs, Symbol("x"))

    def test_times_one(self):
        r = Symbol("r", REAL)
        f = Times(r, r, Real(1))
        f = f.simplify()
        self.assertNotIn(Real(1), f.args())

    @skipIfSolverNotAvailable("z3")
    def test_plus_negatives(self):
        r0 = Symbol("r0", REAL)
        r1 = Symbol("r1", REAL)
        p_1 = Real(1)
        m_1 = Real(-1)
        p_2 = Real(2)
        m_4 = Real(-4)

        # 4 * r0 + (-1) * r1 + 2 - 4
        neg_r1 = Times(m_1, r1)
        m_4_r0 = Times(Real(4), r0)
        expr = Plus(m_4_r0, neg_r1, p_2, m_4)
        res = expr.simplify()
        self.assertValid(Equals(expr, res))
        stack = [res]
        while stack:
            curr = stack.pop()
            if curr.is_plus():
                stack.extend(curr.args())
            elif curr.is_minus():
                stack.extend(curr.args())
            elif curr.is_times():
                stack.extend(curr.args())
            elif curr.is_constant():
                self.assertNotEqual(curr, m_1)
                self.assertNotEqual(curr, p_1)
            elif not curr.is_symbol():
                # unexpected expression type.
                self.assertTrue(False)

    @skipIfSolverNotAvailable("z3")
    def test_sum_all_negatives(self):
        r0 = Symbol("r0", REAL)
        r1 = Symbol("r1", REAL)
        m_1 = Real(-1)

        # -4 * r0 + (-1) * r1
        neg_r1 = Times(m_1, r1)
        m_4_r0 = Times(Real(-4), r0)
        expr = Plus(m_4_r0, neg_r1)
        res = expr.simplify()
        self.assertValid(Equals(expr, res))

    @skipIfSolverNotAvailable("z3")
    def test_plus_algebraic(self):
        from pysmt.constants import Numeral

        env = get_env()
        mgr = env.formula_manager
        r0 = Symbol("r0", REAL)
        p_2 = Real(2)
        m_5 = mgr._Algebraic(Numeral(-5))
        m_3 = mgr._Algebraic(Numeral(-3))

        # r0 + 2 - 5
        expr = Plus(r0, p_2, m_5)
        res = expr.simplify()
        self.assertValid(Equals(expr, res))
        self.assertIn(m_3, res.args())

    @skipIfSolverNotAvailable("z3")
    def test_times_algebraic(self):
        from pysmt.constants import Numeral

        env = get_env()
        mgr = env.formula_manager
        r0 = Symbol("r0", REAL)
        p_2 = Real(2)
        m_5 = mgr._Algebraic(Numeral(-5))
        m_10 = mgr._Algebraic(Numeral(-10))

        # -5 * r0 * 2
        expr = Times(m_5, r0, p_2)
        res = expr.simplify()
        self.assertValid(Equals(expr, res))
        self.assertIn(m_10, res.args())

    def test_and_flattening(self):
        x, y, z = (Symbol(name) for name in "xyz")
        f1 = And(x, y, z)
        f2 = And(x, And(y, z))
        self.assertEqual(f2.simplify(), f1)

    def test_or_flattening(self):
        x, y, z = (Symbol(name) for name in "xyz")
        f1 = Or(x, y, z)
        f2 = Or(x, Or(y, z))
        self.assertEqual(f2.simplify(), f1)

    def test_trivial_false_and(self):
        x, y, z = (Symbol(name) for name in "xyz")
        f = And(x, y, z, Not(x))
        self.assertEqual(f.simplify(), FALSE())

    def test_trivial_true_or(self):
        x, y, z = (Symbol(name) for name in "xyz")
        f = Or(x, y, z, Not(x))
        self.assertEqual(f.simplify(), TRUE())

    def test_trivial_true_times(self):
        x, y, z = (Symbol(name, REAL) for name in "xyz")
        f = Equals(Times(x, z, y), Times(z, y, x))
        self.assertEqual(f.simplify(), TRUE())


if __name__ == "__main__":
    main()



# Copyright 2014 Andrea Micheli and Marco Gario
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

from pysmt.cmd.installers.base import SolverInstaller


class CuddInstaller(SolverInstaller):

    SOLVER = "bdd"

    def __init__(
        self,
        install_dir,
        bindings_dir,
        solver_version,
        mirror_link=None,
        git_version="HEAD",
    ):
        archive_name = "repycudd-%s.tar.gz" % git_version
        native_link = (
            "https://codeload.github.com/pysmt/repycudd/tar.gz/%s" % git_version
        )
        SolverInstaller.__init__(
            self,
            install_dir=install_dir,
            bindings_dir=bindings_dir,
            solver_version=solver_version,
            archive_name=archive_name,
            native_link=native_link,
            mirror_link=mirror_link,
        )
        self.git_version = git_version

    def compile(self):
        # Select the correct Makefile to be used
        makefile = "Makefile"
        if self.architecture == "x86_64":
            makefile = "Makefile_64bit"

        swig = "swig"
        swig_version = SolverInstaller.run("swig -version", get_output=True)
        if "4.0.1" in swig_version or "4.0.0" in swig_version:
            print(
                "WARNING: the BDD solver does not work with Swig4 < 4.0.2. Fallback to Swig3"
            )
            swig = "swig3.0"  # This is the Ubuntu naming of the executable

        import distutils.sysconfig as sysconfig

        PYTHON_INCLUDE_DIR = sysconfig.get_python_inc()
        SolverInstaller.run(
            "make -C %s -f %s PYTHON_INCL=-I%s SWIG=%s"
            % (self.extract_path, makefile, PYTHON_INCLUDE_DIR, swig)
        )

    def move(self):
        SolverInstaller.mv(
            os.path.join(self.extract_path, "repycudd.py"), self.bindings_dir
        )
        SolverInstaller.mv(
            os.path.join(self.extract_path, "_repycudd.so"), self.bindings_dir
        )

    def get_installed_version(self):
        return self.get_installed_version_script(self.bindings_dir, "cudd")



from typing import Dict, List, Tuple, Set, Any
from progpysmt.exceptions import PysmtSyntaxError


class Method:
    def __init__(self, name: str, signature: str) -> None:
        self.name = name
        self.signature = signature

    def __str__(self) -> str:
        return "%s (%s)" % (self.name, self.signature)


class Constraint:
    def __init__(self, var: Dict[str, str], const: Dict[str, Set[str]]) -> None:
        self.table_of_symbols: Dict[str, List[Any]] = {}
        for v in var:
            self.table_of_symbols[v] = None
        for c in const:
            for v in const[c]:
                self.table_of_symbols[v] = v
        self.pbe: Dict[str, Any] = {}


class PSLObject:
    def __init__(self, filename: str) -> None:
        self.filename: str = filename
        self.logics: List[str] = []
        self.type: str = None
        self.pbe: List[Tuple[List[str], str]] = []
        self.func_name: str = None
        self.constraints_syntax: List[str] = []
        self.check: bool = False
        self.solution: str = None
        self.constants: Dict[str, Set[str]] = {}
        self.grammar_interpretation: Dict[str, List[str | Method]] = {}
        self.variables: Dict[str, str] = {}
        self.constraints: List[Constraint] = []
        self.smtlogic: str = ""
        self.methods: Dict[Tuple(str, str), str] = {}

    def format_logics(self, elements: List[str]) -> None:
        res = []
        opened = False
        logic = ""
        args = []
        for e in elements:
            if e == "(":
                if opened:
                    raise PysmtSyntaxError(
                        "Unexpected token %s of command set-logic" % e
                    )
                opened = True
            elif e == ")":
                if not opened:
                    raise PysmtSyntaxError(
                        "Unexpected token %s of command set-logic" % e
                    )
                opened = False
            else:
                if logic == "":
                    logic = e
                else:
                    args.append(e)
            if not opened:
                res.append((logic, args))
                logic = ""
                args = []
        self.logics = res



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['progpysmt',
 'progpysmt.cmd',
 'progpysmt.cmd.installers',
 'progpysmt.smtlib',
 'progpysmt.smtlib.parser',
 'progpysmt.solvers',
 'progpysmt.test',
 'progpysmt.test.smtlib',
 'progpysmt.walkers']

package_data = \
{'': ['*']}

install_requires = \
['cython>=3.0.8,<4.0.0']

setup_kwargs = {
    'name': 'progpysmt',
    'version': '0.1.1',
    'description': 'A Python parser for SMT-LIB, SyGuS and ProgSynth formats.',
    'long_description': '# progpysmt - A SMT-LIB, SyGuS and ProgSynth parser\n\nThis repository is a modification of [pySMT: a Python API for SMT](https://github.com/pysmt/pysmt). The goal of this project is to extend the parser in order to accept the [SyGuS](https://sygus-org.github.io/) format and adapts the output to the [ProgSynth](https://github.com/nathanael-fijalkow/ProgSynth/tree/main) format. \n\n# Usage\nIn order to use the SMT-LIB parser from the original `pySMT` repository, please follow the instructions on their personal github. \nIn order to use the parser on a given file of name `file_name`:\n```Python\nfrom  progpysmt.smtlib.parser  import  ProgSmtLibParser\nfrom  progpysmt.pslobject  import  PSLObject\n\nparser = ProgSmtLibParser()\nwith open(file_name) as f:\n\tpslobject = parser.get_script(f, file_name)\n```   \nThis will create a `PSLObject` that contains all the values required to use ProgSynth.\n\n# Installation\n## From source\nIf you install this from source, you will need Python 3.7 or higher.\n## Poetry\n \n',
    'author': 'Gaëtan MARGUERITTE',
    'author_email': 'gamargueritte@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



#
# This file is part of pySMT.
#
#   Copyright 2014 Andrea Micheli and Marco Gario
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from collections import namedtuple

import pysmt.logics
from pysmt.environment import get_env
from pysmt.shortcuts import (
    Symbol,
    Function,
    Int,
    Real,
    FALSE,
    TRUE,
    And,
    Iff,
    Or,
    Not,
    Implies,
    Ite,
    LT,
    LE,
    GT,
    GE,
    Times,
    Pow,
    Equals,
    Plus,
    Minus,
    Div,
    ToReal,
    ForAll,
    Exists,
    BV,
    SBV,
    BVOne,
    BVZero,
    BVNot,
    BVAnd,
    BVOr,
    BVXor,
    BVConcat,
    BVExtract,
    BVULT,
    BVUGT,
    BVULE,
    BVUGE,
    BVSGE,
    BVNeg,
    BVAdd,
    BVMul,
    BVUDiv,
    BVURem,
    BVSub,
    BVLShl,
    BVLShr,
    BVRol,
    BVRor,
    BVZExt,
    BVSExt,
    BVSub,
    BVComp,
    BVAShr,
    BVSLE,
    BVSLT,
    BVSGT,
    BVSGE,
    BVSDiv,
    BVSRem,
    String,
    StrCharAt,
    StrConcat,
    StrContains,
    StrIndexOf,
    StrLength,
    StrPrefixOf,
    StrReplace,
    StrSubstr,
    StrSuffixOf,
    StrToInt,
    IntToStr,
    BVToNatural,
    Store,
    Select,
    Array,
    Type,
)

from pysmt.typing import REAL, BOOL, INT, BV8, BV16, BVType, ARRAY_INT_INT
from pysmt.typing import FunctionType, ArrayType, STRING
from pysmt.constants import Fraction


Example = namedtuple("Example", ["hr", "expr", "is_valid", "is_sat", "logic"])

SExample = namedtuple("SimpleExample", ["expr", "is_valid", "is_sat", "logic"])


def get_full_example_formulae(environment=None):
    """Return a list of Examples using the given environment."""

    if environment is None:
        environment = get_env()

    with environment:
        x = Symbol("x", BOOL)
        y = Symbol("y", BOOL)
        p = Symbol("p", INT)
        q = Symbol("q", INT)
        r = Symbol("r", REAL)
        s = Symbol("s", REAL)
        aii = Symbol("aii", ARRAY_INT_INT)
        ari = Symbol("ari", ArrayType(REAL, INT))
        arb = Symbol("arb", ArrayType(REAL, BV8))
        abb = Symbol("abb", ArrayType(BV8, BV8))
        nested_a = Symbol("a_arb_aii", ArrayType(ArrayType(REAL, BV8), ARRAY_INT_INT))

        rf = Symbol("rf", FunctionType(REAL, [REAL, REAL]))
        rg = Symbol("rg", FunctionType(REAL, [REAL]))

        ih = Symbol("ih", FunctionType(INT, [REAL, INT]))

        bf = Symbol("bf", FunctionType(BOOL, [BOOL]))
        bg = Symbol("bg", FunctionType(BOOL, [BOOL]))

        bv3 = Symbol("bv3", BVType(3))
        bv8 = Symbol("bv8", BV8)
        bv16 = Symbol("bv16", BV16)

        unary_sort = Type("S", 1)
        tmgr = environment.type_manager
        unary_sort_bool = tmgr.get_type_instance(unary_sort, BOOL)
        usb1 = Symbol("usb1", unary_sort_bool)

        str1 = Symbol("str1", STRING)

        result = [
            # Formula, is_valid, is_sat, is_qf
            Example(
                hr="(x & y)",
                expr=And(x, y),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_BOOL,
            ),
            Example(
                hr="(x <-> y)",
                expr=Iff(x, y),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_BOOL,
            ),
            Example(
                hr="((x | y) & (! (x | y)))",
                expr=And(Or(x, y), Not(Or(x, y))),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.QF_BOOL,
            ),
            Example(
                hr="(x & (! y))",
                expr=And(x, Not(y)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_BOOL,
            ),
            Example(
                hr="(False -> True)",
                expr=Implies(FALSE(), TRUE()),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_BOOL,
            ),
            Example(
                hr="((x | y) & (! (x | y)))",
                expr=And(Or(x, y), Not(Or(x, y))),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.QF_BOOL,
            ),
            #
            #  LIA
            #
            Example(
                hr="((q < p) & (x -> y))",
                expr=And(GT(p, q), Implies(x, y)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_IDL,
            ),
            Example(
                hr="(((p + q) = 5) & (q < p))",
                expr=And(Equals(Plus(p, q), Int(5)), GT(p, q)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_LIA,
            ),
            Example(
                hr="((q <= p) | (p <= q))",
                expr=Or(GE(p, q), LE(p, q)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_IDL,
            ),
            Example(
                hr="(! (p < (q * 2)))",
                expr=Not(LT(p, Times(q, Int(2)))),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_LIA,
            ),
            Example(
                hr="(p < (p - (5 - 2)))",
                expr=GT(Minus(p, Minus(Int(5), Int(2))), p),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.QF_IDL,
            ),
            Example(
                hr="((x ? 7 : ((p + -1) * 3)) = q)",
                expr=Equals(Ite(x, Int(7), Times(Plus(p, Int(-1)), Int(3))), q),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_LIA,
            ),
            Example(
                hr="(p < (q + 1))",
                expr=LT(p, Plus(q, Int(1))),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_LIA,
            ),
            #
            # LRA
            #
            Example(
                hr="((s < r) & (x -> y))",
                expr=And(GT(r, s), Implies(x, y)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_RDL,
            ),
            Example(
                hr="(((r + s) = 28/5) & (s < r))",
                expr=And(Equals(Plus(r, s), Real(Fraction("5.6"))), GT(r, s)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_LRA,
            ),
            Example(
                hr="((s <= r) | (r <= s))",
                expr=Or(GE(r, s), LE(r, s)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_RDL,
            ),
            Example(
                hr="(! ((r * 2.0) < (s * 2.0)))",
                expr=Not(LT(Div(r, Real((1, 2))), Times(s, Real(2)))),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_LRA,
            ),
            Example(
                hr="(! (r < (r - (5.0 - 2.0))))",
                expr=Not(GT(Minus(r, Minus(Real(5), Real(2))), r)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_RDL,
            ),
            Example(
                hr="((x ? 7.0 : ((s + -1.0) * 3.0)) = r)",
                expr=Equals(Ite(x, Real(7), Times(Plus(s, Real(-1)), Real(3))), r),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_LRA,
            ),
            #
            # EUF
            #
            Example(
                hr="(bf(x) <-> bg(x))",
                expr=Iff(Function(bf, (x,)), Function(bg, (x,))),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_UF,
            ),
            Example(
                hr="(rf(5.0, rg(r)) = 0.0)",
                expr=Equals(Function(rf, (Real(5), Function(rg, (r,)))), Real(0)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_UFLRA,
            ),
            Example(
                hr="((rg(r) = (5.0 + 2.0)) <-> (rg(r) = 7.0))",
                expr=Iff(
                    Equals(Function(rg, [r]), Plus(Real(5), Real(2))),
                    Equals(Function(rg, [r]), Real(7)),
                ),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_UFLRA,
            ),
            Example(
                hr="((r = (s + 1.0)) & (rg(s) = 5.0) & (rg((r - 1.0)) = 7.0))",
                expr=And(
                    [
                        Equals(r, Plus(s, Real(1))),
                        Equals(Function(rg, [s]), Real(5)),
                        Equals(Function(rg, [Minus(r, Real(1))]), Real(7)),
                    ]
                ),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.QF_UFLRA,
            ),
            #
            # BV
            #
            Example(
                hr="((1_32 & 0_32) = 0_32)",
                expr=Equals(BVAnd(BVOne(32), BVZero(32)), BVZero(32)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((! 2_3) = 5_3)",
                expr=Equals(BVNot(BV("010")), BV("101")),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((! bv3) = 5_3)",
                expr=Equals(BVNot(bv3), BV("101")),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((7_3 xor 0_3) = 0_3)",
                expr=Equals(BVXor(BV("111"), BV("000")), BV("000")),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((7_3 xor bv3) = (6_3 xor bv3))",
                expr=Equals(BVXor(BV("111"), bv3), BVXor(BV("110"), bv3)),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((bv8::bv8) u< 0_16)",
                expr=BVULT(BVConcat(bv8, bv8), BVZero(16)),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((bv8::bv8) u< (bv8::9_8))",
                expr=BVULT(BVConcat(bv8, bv8), BVConcat(bv8, BV(9, 8))),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="(1_32[0:7] = 1_8)",
                expr=Equals(BVExtract(BVOne(32), end=7), BVOne(8)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="(0_8 u< (((bv8 + 1_8) * 5_8) u/ 5_8))",
                expr=BVUGT(
                    BVUDiv(BVMul(BVAdd(bv8, BVOne(8)), BV(5, width=8)), BV(5, width=8)),
                    BVZero(8),
                ),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="(0_16 u<= bv16)",
                expr=BVUGE(bv16, BVZero(16)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="(0_16 s<= bv16)",
                expr=BVSGE(bv16, BVZero(16)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((0_32 u< (5_32 u% 2_32)) & ((5_32 u% 2_32) u<= 1_32))",
                expr=And(
                    BVUGT(BVURem(BV(5, width=32), BV(2, width=32)), BVZero(32)),
                    BVULE(BVURem(BV(5, width=32), BV(2, width=32)), BVOne(32)),
                ),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((((1_32 + (- 1_32)) << 1_32) >> 1_32) = 1_32)",
                expr=Equals(
                    BVLShr(BVLShl(BVAdd(BVOne(32), BVNeg(BVOne(32))), 1), 1), BVOne(32)
                ),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((1_32 - 1_32) = 0_32)",
                expr=Equals(BVSub(BVOne(32), BVOne(32)), BVZero(32)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            # Rotations
            Example(
                hr="(((1_32 ROL 1) ROR 1) = 1_32)",
                expr=Equals(BVRor(BVRol(BVOne(32), 1), 1), BVOne(32)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((bv16 ROL 1) = (bv16 ROR 2))",
                expr=Equals(BVRol(bv16, 1), BVRor(bv16, 2)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            # Extensions
            Example(
                hr="((0_5 ZEXT 11) = (0_1 SEXT 15))",
                expr=Equals(BVZExt(BVZero(5), 11), BVSExt(BVZero(1), 15)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((bv8 ZEXT 19) = (bv16 SEXT 11))",
                expr=Equals(BVZExt(bv8, 19), BVSExt(bv16, 11)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((bv16 - bv16) = 0_16)",
                expr=Equals(BVSub(bv16, bv16), BVZero(16)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((bv16 - bv16)[0:7] = bv8)",
                expr=Equals(BVExtract(BVSub(bv16, bv16), 0, 7), bv8),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((bv16[0:7] bvcomp bv8) = 1_1)",
                expr=Equals(BVComp(BVExtract(bv16, 0, 7), bv8), BVOne(1)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((bv16 bvcomp bv16) = 0_1)",
                expr=Equals(BVComp(bv16, bv16), BVZero(1)),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="(bv16 s< bv16)",
                expr=BVSLT(bv16, bv16),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="(bv16 s< 0_16)",
                expr=BVSLT(bv16, BVZero(16)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((bv16 s< 0_16) | (0_16 s<= bv16))",
                expr=Or(BVSGT(BVZero(16), bv16), BVSGE(bv16, BVZero(16))),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="(bv16 u< bv16)",
                expr=BVULT(bv16, bv16),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="(bv16 u< 0_16)",
                expr=BVULT(bv16, BVZero(16)),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((bv16 | 0_16) = bv16)",
                expr=Equals(BVOr(bv16, BVZero(16)), bv16),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((bv16 | 5_16) = bv16)",
                expr=Equals(BVOr(bv16, BV(5, 16)), bv16),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((bv16 & 0_16) = 0_16)",
                expr=Equals(BVAnd(bv16, BVZero(16)), BVZero(16)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((bv16 & 7_16) = 0_16)",
                expr=Equals(BVAnd(bv16, BV(7, 16)), BVZero(16)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((0_16 s< bv16) & ((bv16 s/ 65535_16) s< 0_16))",
                expr=And(
                    BVSLT(BVZero(16), bv16),
                    BVSLT(BVSDiv(bv16, SBV(-1, 16)), BVZero(16)),
                ),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((0_16 s< bv16) & ((bv16 s% 1_16) s< 0_16))",
                expr=And(
                    BVSLT(BVZero(16), bv16), BVSLT(BVSRem(bv16, BVOne(16)), BVZero(16))
                ),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((bv16 u% 1_16) = 0_16)",
                expr=Equals(BVURem(bv16, BVOne(16)), BVZero(16)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((bv16 u% bv16) = 0_16)",
                expr=Equals(BVURem(bv16, bv16), BVZero(16)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((bv16 s% 1_16) = 0_16)",
                expr=Equals(BVSRem(bv16, BVOne(16)), BVZero(16)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((bv16 s% bv16) = 0_16)",
                expr=Equals(BVSRem(bv16, bv16), BVZero(16)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((bv16 s% (- 1_16)) = 0_16)",
                expr=Equals(BVSRem(bv16, BVNeg(BVOne(16))), BVZero(16)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="(bv16 s< (- bv16))",
                expr=BVSGT(BVNeg(bv16), bv16),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((bv16 a>> 0_16) = bv16)",
                expr=Equals(BVAShr(bv16, BVZero(16)), bv16),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="((0_16 s<= bv16) & ((bv16 a>> 1_16) = (bv16 >> 1_16)))",
                expr=And(
                    BVSLE(BVZero(16), bv16),
                    Equals(BVAShr(bv16, BVOne(16)), BVLShr(bv16, BVOne(16))),
                ),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_BV,
            ),
            Example(
                hr="(bv2nat(1_8) = 1)",
                expr=Equals(BVToNatural(BVOne(8)), Int(1)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_AUFBVLIRA,
            ),
            #
            # Quantification
            #
            Example(
                hr="(forall y . (x -> y))",
                expr=ForAll([y], Implies(x, y)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.BOOL,
            ),
            Example(
                hr="(forall p, q . ((p + q) = 0))",
                expr=ForAll([p, q], Equals(Plus(p, q), Int(0))),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.LIA,
            ),
            Example(
                hr="(forall r, s . (((0.0 < r) & (0.0 < s)) -> ((r - s) < r)))",
                expr=ForAll(
                    [r, s],
                    Implies(And(GT(r, Real(0)), GT(s, Real(0))), (LT(Minus(r, s), r))),
                ),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.LRA,
            ),
            Example(
                hr="(exists x, y . (x -> y))",
                expr=Exists([x, y], Implies(x, y)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.BOOL,
            ),
            Example(
                hr="(exists p, q . ((p + q) = 0))",
                expr=Exists([p, q], Equals(Plus(p, q), Int(0))),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.LIA,
            ),
            Example(
                hr="(exists r . (forall s . (r < (r - s))))",
                expr=Exists([r], ForAll([s], GT(Minus(r, s), r))),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.LRA,
            ),
            Example(
                hr="(forall r . (exists s . (r < (r - s))))",
                expr=ForAll([r], Exists([s], GT(Minus(r, s), r))),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.LRA,
            ),
            Example(
                hr="(x & (forall r . ((r + s) = 5.0)))",
                expr=And(x, ForAll([r], Equals(Plus(r, s), Real(5)))),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.LRA,
            ),
            Example(
                hr="(exists x . ((x <-> (5.0 < s)) & (s < 3.0)))",
                expr=Exists([x], (And(Iff(x, GT(s, Real(5))), LT(s, Real(3))))),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.LRA,
            ),
            #
            # UFLIRA
            #
            Example(
                hr="((p < ih(r, q)) & (x -> y))",
                expr=And(GT(Function(ih, (r, q)), p), Implies(x, y)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_UFLIRA,
            ),
            Example(
                hr="(((p - 3) = q) -> ((p < ih(r, (q + 3))) | (ih(r, p) <= p)))",
                expr=Implies(
                    Equals(Minus(p, Int(3)), q),
                    Or(
                        GT(Function(ih, (r, Plus(q, Int(3)))), p),
                        LE(Function(ih, (r, p)), p),
                    ),
                ),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_UFLIRA,
            ),
            Example(
                hr="(((ToReal((p - 3)) = r) & (ToReal(q) = r)) -> ((p < ih(ToReal((p - 3)), (q + 3))) | (ih(r, p) <= p)))",
                expr=Implies(
                    And(Equals(ToReal(Minus(p, Int(3))), r), Equals(ToReal(q), r)),
                    Or(
                        GT(
                            Function(ih, (ToReal(Minus(p, Int(3))), Plus(q, Int(3)))), p
                        ),
                        LE(Function(ih, (r, p)), p),
                    ),
                ),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_UFLIRA,
            ),
            Example(
                hr="(! (((ToReal((p - 3)) = r) & (ToReal(q) = r)) -> ((p < ih(ToReal((p - 3)), (q + 3))) | (ih(r, p) <= p))))",
                expr=Not(
                    Implies(
                        And(Equals(ToReal(Minus(p, Int(3))), r), Equals(ToReal(q), r)),
                        Or(
                            GT(
                                Function(
                                    ih, (ToReal(Minus(p, Int(3))), Plus(q, Int(3)))
                                ),
                                p,
                            ),
                            LE(Function(ih, (r, p)), p),
                        ),
                    )
                ),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.QF_UFLIRA,
            ),
            #
            # STR
            #
            Example(
                hr='("mystr" = str1)',
                expr=Equals(String("mystr"), str1),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_SLIA,
            ),
            Example(
                hr='((5 < str.len(str1)) & ("mystr" = str1))',
                expr=And(LT(Int(5), StrLength(str1)), Equals(String("mystr"), str1)),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.QF_SLIA,
            ),
            Example(
                hr='((5 = str.len(str1)) & (str.++("my", "str") = str1))',
                expr=And(
                    Equals(Int(5), StrLength(str1)),
                    Equals(StrConcat(String("my"), String("str")), str1),
                ),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_SLIA,
            ),
            Example(
                hr='(str.at("mystr", 1) = "y")',
                expr=Equals(StrCharAt(String("mystr"), Int(1)), String("y")),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_SLIA,
            ),
            Example(
                hr='str.contains("mystr", "my")',
                expr=StrContains(String("mystr"), String("my")),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_SLIA,
            ),
            Example(
                hr='(str.indexof("mystr", "str", 1) = 2)',
                expr=Equals(StrIndexOf(String("mystr"), String("str"), Int(1)), Int(2)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_SLIA,
            ),
            Example(
                hr='(str.indexof(str1, "str", 1) = 2)',
                expr=Equals(StrIndexOf(str1, String("str"), Int(1)), Int(2)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_SLIA,
            ),
            Example(
                hr='(str.replace("mystr", "str", "my") = "mymy")',
                expr=Equals(
                    StrReplace(String("mystr"), String("str"), String("my")),
                    String("mymy"),
                ),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_SLIA,
            ),
            Example(
                hr='(str.substr("mystr", 2, 4) = "st")',
                expr=Equals(StrSubstr(String("mystr"), Int(2), Int(4)), String("st")),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.QF_SLIA,
            ),
            Example(
                hr='str.prefixof("my", "mystr")',
                expr=StrPrefixOf(String("my"), String("mystr")),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_SLIA,
            ),
            Example(
                hr='str.suffixof("str", "mystr")',
                expr=StrSuffixOf(String("str"), String("mystr")),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_SLIA,
            ),
            Example(
                hr='(int.to.str(9) = "9")',
                expr=Equals(IntToStr(Int(9)), String("9")),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_SLIA,
            ),
            Example(
                hr='(str.to.int("9") = 9)',
                expr=Equals(StrToInt(String("9")), Int(9)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_SLIA,
            ),
            Example(
                hr="('Did you know that any string works? #yolo' & '10' & '|#somesolverskeepthe||' & ' ')"
                "",
                expr=And(
                    Symbol("Did you know that any string works? #yolo"),
                    Symbol("10"),
                    Symbol("|#somesolverskeepthe||"),
                    Symbol(" "),
                ),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_BOOL,
            ),
            #
            # Arrays
            #
            Example(
                hr="((q = 0) -> (aii[0 := 0] = aii[0 := q]))",
                expr=Implies(
                    Equals(q, Int(0)),
                    Equals(Store(aii, Int(0), Int(0)), Store(aii, Int(0), q)),
                ),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_ALIA,
            ),
            Example(
                hr="(aii[0 := 0][0] = 0)",
                expr=Equals(Select(Store(aii, Int(0), Int(0)), Int(0)), Int(0)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_ALIA,
            ),
            Example(
                hr="((Array{Int, Int}(0)[1 := 1] = aii) & (aii[1] = 0))",
                expr=And(
                    Equals(Array(INT, Int(0), {Int(1): Int(1)}), aii),
                    Equals(Select(aii, Int(1)), Int(0)),
                ),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.get_logic_by_name("QF_ALIA*"),
            ),
            Example(
                hr="((Array{Int, Int}(0)[1 := 3] = aii) & (aii[1] = 3))",
                expr=And(
                    Equals(Array(INT, Int(0), {Int(1): Int(3)}), aii),
                    Equals(Select(aii, Int(1)), Int(3)),
                ),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.get_logic_by_name("QF_ALIA*"),
            ),
            Example(
                hr="((Array{Real, Int}(10) = ari) & (ari[6/5] = 0))",
                expr=And(
                    Equals(Array(REAL, Int(10)), ari),
                    Equals(Select(ari, Real((6, 5))), Int(0)),
                ),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.get_logic_by_name("QF_AUFBVLIRA*"),
            ),
            Example(
                hr="((Array{Real, Int}(0)[1.0 := 10][2.0 := 20][3.0 := 30][4.0 := 40] = ari) & (! ((ari[0.0] = 0) & (ari[1.0] = 10) & (ari[2.0] = 20) & (ari[3.0] = 30) & (ari[4.0] = 40))))",
                expr=And(
                    Equals(
                        Array(
                            REAL,
                            Int(0),
                            {
                                Real(1): Int(10),
                                Real(2): Int(20),
                                Real(3): Int(30),
                                Real(4): Int(40),
                            },
                        ),
                        ari,
                    ),
                    Not(
                        And(
                            Equals(Select(ari, Real(0)), Int(0)),
                            Equals(Select(ari, Real(1)), Int(10)),
                            Equals(Select(ari, Real(2)), Int(20)),
                            Equals(Select(ari, Real(3)), Int(30)),
                            Equals(Select(ari, Real(4)), Int(40)),
                        )
                    ),
                ),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.get_logic_by_name("QF_AUFBVLIRA*"),
            ),
            Example(
                hr="((Array{Real, Int}(0)[1.0 := 10][2.0 := 20][3.0 := 30][4.0 := 40][5.0 := 50] = ari) & (! ((ari[0.0] = 0) & (ari[1.0] = 10) & (ari[2.0] = 20) & (ari[3.0] = 30) & (ari[4.0] = 40) & (ari[5.0] = 50))))",
                expr=And(
                    Equals(
                        Array(
                            REAL,
                            Int(0),
                            {
                                Real(1): Int(10),
                                Real(2): Int(20),
                                Real(3): Int(30),
                                Real(4): Int(40),
                                Real(5): Int(50),
                            },
                        ),
                        ari,
                    ),
                    Not(
                        And(
                            Equals(Select(ari, Real(0)), Int(0)),
                            Equals(Select(ari, Real(1)), Int(10)),
                            Equals(Select(ari, Real(2)), Int(20)),
                            Equals(Select(ari, Real(3)), Int(30)),
                            Equals(Select(ari, Real(4)), Int(40)),
                            Equals(Select(ari, Real(5)), Int(50)),
                        )
                    ),
                ),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.get_logic_by_name("QF_AUFBVLIRA*"),
            ),
            Example(
                hr="((a_arb_aii = Array{Array{Real, BV{8}}, Array{Int, Int}}(Array{Int, Int}(7))) -> (a_arb_aii[arb][42] = 7))",
                expr=Implies(
                    Equals(nested_a, Array(ArrayType(REAL, BV8), Array(INT, Int(7)))),
                    Equals(Select(Select(nested_a, arb), Int(42)), Int(7)),
                ),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.get_logic_by_name("QF_AUFBVLIRA*"),
            ),
            Example(
                hr="(abb[bv8 := y_][bv8 := z_] = abb[bv8 := z_])",
                expr=Equals(
                    Store(Store(abb, bv8, Symbol("y_", BV8)), bv8, Symbol("z_", BV8)),
                    Store(abb, bv8, Symbol("z_", BV8)),
                ),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_ABV,
            ),
            Example(
                hr="((r / s) = (r * s))",
                expr=Equals(Div(r, s), Times(r, s)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_NRA,
            ),
            Example(
                hr="(2.0 = (r * r))",
                expr=Equals(Real(2), Times(r, r)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_NRA,
            ),
            Example(
                hr="((p ^ 2) = 0)",
                expr=Equals(Pow(p, Int(2)), Int(0)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_NIA,
            ),
            Example(
                hr="((r ^ 2.0) = 0.0)",
                expr=Equals(Pow(r, Real(2)), Real(0)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_NRA,
            ),
            Example(
                hr="((r * r * r) = 25.0)",
                expr=Equals(Times(r, r, r), Real(25)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_NRA,
            ),
            Example(
                hr="((5.0 * r * 5.0) = 25.0)",
                expr=Equals(Times(Real(5), r, Real(5)), Real(25)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_LRA,
            ),
            Example(
                hr="((p * p * p) = 25)",
                expr=Equals(Times(p, p, p), Int(25)),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.QF_NIA,
            ),
            Example(
                hr="((5 * p * 5) = 25)",
                expr=Equals(Times(Int(5), p, Int(5)), Int(25)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_LIA,
            ),
            Example(
                hr="(((1 - 1) * p * 1) = 0)",
                expr=Equals(Times(Minus(Int(1), Int(1)), p, Int(1)), Int(0)),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_LIA,
            ),
            # Huge Fractions:
            Example(
                hr="((r * 1606938044258990275541962092341162602522202993782792835301376/7) = -20480000000000000000000000.0)",
                expr=Equals(Times(r, Real(Fraction(2**200, 7))), Real(-(200**11))),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_LRA,
            ),
            Example(
                hr="(((r + 5.0 + s) * (s + 2.0 + r)) = 0.0)",
                expr=Equals(Times(Plus(r, Real(5), s), Plus(s, Real(2), r)), Real(0)),
                is_valid=False,
                is_sat=True,
                logic=pysmt.logics.QF_NRA,
            ),
            Example(
                hr="(((p + 5 + q) * (p - (q - 5))) = ((p * p) + (10 * p) + 25 + (-1 * q * q)))",
                expr=Equals(
                    Times(Plus(p, Int(5), q), Minus(p, Minus(q, Int(5)))),
                    Plus(Times(p, p), Times(Int(10), p), Int(25), Times(Int(-1), q, q)),
                ),
                is_valid=True,
                is_sat=True,
                logic=pysmt.logics.QF_NIA,
            ),
            # Sorts
            Example(
                hr="(! (usb1 = usb1))",
                expr=Not(Equals(usb1, usb1)),
                is_valid=False,
                is_sat=False,
                logic=pysmt.logics.QF_BOOLt,
            ),
        ]
    return result


def get_example_formulae(environment=None):
    """Generates a stream of SExample using the given environment."""
    for (hr, expr, is_valid, is_sat, logic) in get_full_example_formulae(environment):
        yield SExample(expr, is_valid, is_sat, logic)


def get_str_example_formulae(environment=None):
    """Returns a triple from each Example."""
    for (hr, expr, is_valid, is_sat, logic) in get_full_example_formulae(environment):
        yield (hr, expr, logic)



# Copyright 2014 Andrea Micheli and Marco Gario
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys
import glob

from pysmt.cmd.installers.base import SolverInstaller, TemporaryPath


class YicesInstaller(SolverInstaller):

    SOLVER = "yices"

    def __init__(
        self,
        install_dir,
        bindings_dir,
        solver_version,
        mirror_link=None,
        yicespy_version="HEAD",
    ):

        archive_name = "Yices-%s.tar.gz" % (solver_version)
        native_link = "https://github.com/SRI-CSL/yices2/archive/{archive_name}"
        SolverInstaller.__init__(
            self,
            install_dir=install_dir,
            bindings_dir=bindings_dir,
            solver_version=solver_version,
            archive_name=archive_name,
            native_link=native_link,
            mirror_link=mirror_link,
        )

        self.extract_path = os.path.join(
            self.base_dir, "yices2-Yices-%s" % self.solver_version
        )
        self.yices_path = os.path.join(self.bindings_dir, "yices_bin")
        self.yicespy_git_version = yicespy_version

    def install_yicespy(self):
        yicespy_git_version = self.yicespy_git_version
        yicespy_base_name = "yicespy"
        yicespy_archive_name = "%s.tar.gz" % yicespy_base_name
        yicespy_archive = os.path.join(self.base_dir, yicespy_archive_name)
        yicespy_dir_path = os.path.join(
            self.base_dir, yicespy_base_name + "-" + yicespy_git_version
        )

        yicespy_download_link = (
            "https://codeload.github.com/pysmt/yicespy/tar.gz/%s"
            % (yicespy_git_version)
        )
        SolverInstaller.do_download(yicespy_download_link, yicespy_archive)

        SolverInstaller.clean_dir(yicespy_dir_path)

        SolverInstaller.untar(yicespy_archive, self.base_dir)
        # Build yicespy
        SolverInstaller.run_python(
            "setup.py --yices-dir=%s -- build_ext bdist_wheel --dist-dir=%s "
            % (self.yices_path, self.base_dir),
            directory=yicespy_dir_path,
        )
        wheel_file = glob.glob(os.path.join(self.base_dir, "yicespy") + "*.whl")[0]
        SolverInstaller.unzip(wheel_file, self.bindings_dir)

    def compile(self):
        # Prepare an empty folder for installing yices
        SolverInstaller.clean_dir(self.yices_path)

        SolverInstaller.run("autoconf", directory=self.extract_path)

        SolverInstaller.run(
            "bash configure --prefix %s" % self.yices_path, directory=self.extract_path
        )
        SolverInstaller.run("make", directory=self.extract_path)
        SolverInstaller.run("make install", directory=self.extract_path)

        self.install_yicespy()

    def get_installed_version(self):
        return self.get_installed_version_script(self.bindings_dir, "yices")
